#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""RPC handlers not included in TracXmlRpcPlugin and used to 
implement some data providers supporting Google Visualization API
protocol.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

__all__ = 'TimelineRPC', 'ReportRPC', 'VersionControlRPC'

from trac.core import implements, ExtensionPoint, Component
from trac.mimeview.api import Mimeview, Context
from trac.ticket.report import ReportModule
from trac.ticket.query import Query, QuerySyntaxError
from trac.timeline.api import ITimelineEventProvider
from trac.timeline.web_ui import TimelineModule
from trac.util.datefmt import _epoc
from trac.util.translation import _
from trac.versioncontrol.api import RepositoryManager, NoSuchNode, \
                                    NoSuchChangeset
from trac.versioncontrol.web_ui.browser import CHUNK_SIZE 
from trac.web.href import Href

from datetime import datetime, date, time
from itertools import imap
from os.path import splitext
from tracrpc.api import IXMLRPCHandler
import types
from urlparse import urlunparse, urlparse
import xmlrpclib

from util import get_column_desc, rpc_to_datetime, rpc_opt_sigs, \
                  compile_pattern

__metaclass__ = type

#--------------------------------------------------
#   Event Timeline RPC
#--------------------------------------------------

class TimelineRPC(Component):
    r""" An interface to Trac's timeline module.
    """ 
    implements(IXMLRPCHandler)
    sources = ExtensionPoint(ITimelineEventProvider)

    def __init__(self):
        self._event_data = TimelineModule(self.env)._event_data
    
    # IXMLRPCHandler methods
    def xmlrpc_namespace(self):
        return 'timeline'
    
    def xmlrpc_methods(self):
        yield ('TIMELINE_VIEW', 
                ((list, xmlrpclib.DateTime, xmlrpclib.DateTime, list),
                 (list, xmlrpclib.DateTime, xmlrpclib.DateTime), 
                 (list, xmlrpclib.DateTime), 
                 (list, )), 
                 self.getEvents)
        yield ('TIMELINE_VIEW', ((list,),), self.getEventFilters)

    STMT = "(kind, author, date, dateuid, " \
            "unicode(render('url', ctx)).encode('utf-8', 'replace'), " \
            "unicode(render('title', ctx)).encode('utf-8', 'replace'), " \
            "unicode(render('description', ctx)).encode('utf-8', 'replace'))"
    STMT = compile(STMT, '<string>', 'eval')
    
    # Exported methods
    def getEvents(self, req, start=None, stop=None, filters=None):
        r"""Retrieve events taking place in a time window.
        
        @param start    initial date in time interval
        @param stop     last date in time interval
        @param filter   is a list of the enabled filters, 
                        each item being the name (first item) in the 
                        tuples returned by 
                        `ITimelineEventProvider.get_timeline_filters`. 
                        If none is specified then all available 
                        filters will be used.
        @return         a list of events in the time range given by 
                        the `start` and `stop` parameters. Each item 
                        is represented as a tuple of the form 
                        (kind, author, date, timestamp, url, title, desc) 
                        representing information about each event.
        
        
        Note: In order to retrieve all events related to ticket 
              changes (e.g. attachments) you need to set
              timeline.ticket_show_details option in trac.ini to true.
        """
        if start is None:
            start = _epoc
        else:
            start = rpc_to_datetime(start, req)
        if stop is None:
            stop = datetime.now(req.tz)
        else:
            stop = rpc_to_datetime(stop, req)
        if filters is None:
            filters = list(f for f, _ in self.getEventFilters(req))
        ctx = Context.from_request(req, absurls=True)
        globs = dict(ctx=ctx)
        return list(eval(self.STMT, globs, self._event_data(p, e)) \
                                for p in self.sources \
                                for e in p.get_timeline_events(req, \
                                            start, stop, filters))
    
    def getEventFilters(self, req):
        r"""Return a list of the filters available to retrieve events 
        provided  by the timeline module. The data returned for each 
        filter is a binary tuple containing the filter 
        name as well as its display name.
        """
        fdata = dict(fi[:2] for p in self.sources \
                        for fi in p.get_timeline_filters(req))
        return fdata.iteritems()

#--------------------------------------------------
#   Ticket Report RPC
#--------------------------------------------------

class ReportRPC(Component):
    r""" An interface to Trac's report module.
    """ 
    implements(IXMLRPCHandler)
    
    def __init__(self):
      self.repmdl = ReportModule(self.env)
    
    # IXMLRPCHandler methods
    def xmlrpc_namespace(self):
        return 'report'
    
    def xmlrpc_methods(self):
        yield ('REPORT_VIEW', ((list,),), self.getAll)
        yield ('REPORT_VIEW', ((dict, int),), self.get)
        yield ('REPORT_VIEW', ((dict, int),), self.execute)
        yield ('REPORT_VIEW', ((list, int),), self.enum_columns)

    # Exported methods
    def getAll(self, req):
        r"""Return a list containing the IDs of all available reports.
        """
        # Copycat ! I really did nothing
        db = self.env.get_db_cnx()
        sql = ("SELECT id FROM report ORDER BY id")
        cursor = db.cursor()
        try:
          cursor.execute(sql)
          result = cursor.fetchall() or []
          return (x[0] for x in result)
        finally:
          cursor.close()
    
    def get(self, req, id):
        r"""Return information about an specific report as a dict 
        containing the following fields.
        
        - id :            the report ID.
        - title:          the report title.
        - description:    the report description.
        - query:          the query string used to select the tickets 
                          to be included in this report
        """
        sql = "SELECT id,title,query,description from report " \
                 "WHERE id=%s" % (id,)
        db = self.env.get_db_cnx()
        cursor = db.cursor()
        try:
          cursor.execute(sql)
          for report_info in cursor:
              return dict(zip(['id','title','query','description'], report_info))
          else:
              return None
        finally:
          cursor.close()
    
    def _execute_sql(self, req, id, sql, limit=0):
        r"""Execute a SQL report and return no more than `limit` rows 
        (or all rows if limit == 0).
        """
        repmdl = self.repmdl
        db = self.env.get_db_cnx()
        try:
          args = repmdl.get_var_args(req)
        except ValueError,e:
          raise ValueError(_('Report failed: %(error)s', error=e))
        try:
            try:
              # Paginated exec (>=0.11)
              exec_proc = repmdl.execute_paginated_report
              kwargs = dict(limit=limit)
            except AttributeError:
              # Legacy exec (<=0.10)
              exec_proc = repmdl.execute_report
              kwargs = {}
            return exec_proc(req, db, id, sql, args, **kwargs)[:2]
        except Exception, e:
            db.rollback()
            raise 
    
    def execute(self, req, id):
        r"""Execute a Trac report.
        
        @param id     the report ID.
        @return       a list containing the data provided by the 
                      target report.
        @throws       `NotImplementedError` if the report definition 
                      consists of saved custom query specified 
                      using a URL.
        @throws       `QuerySyntaxError` if the report definition 
                      consists of a `TracQuery` containing syntax errors.
        @throws       `Exception` in case of detecting any other error.
        """
        sql = self.get(req, id)['query']
        query = ''.join([line.strip() for line in sql.splitlines()])
        if query and (query[0] == '?' or query.startswith('query:?')):
          raise NotImplementedError('Saved custom queries specified ' \
                                  'using URLs are not supported.')
        elif query.startswith('query:'):
          query = Query.from_string(self.env, query[6:], report=id)
          server_url = urlparse(req.base_url)
          server_href = Href(urlunparse((server_url.scheme, \
                                        server_url.netloc, \
                                        '', '', '', '')))
          def rel2abs(row):
            """Turn relative value in 'href' into absolute URLs."""
            self.log.debug('IG: Query Row %s', row)
            row['href'] = server_href(row['href'])
            return row
            
          return imap(rel2abs, query.execute(req))
        else:
          cols, results = self._execute_sql(req, id, sql)
          return (dict(zip(cols, list(row))) for row in results)
    
    def _sql_cursor(self, req, db, id, sql, args, limit=0, offset=0):
      r"""Retrieve a cursor to access the data returned by a SQL 
      report.
      """
      # Copycat ! ReportModule.execute_paginated_report
      # I didnt want to but I had no other choice :(
      repmdl = self.repmdl
      sql, args = repmdl.sql_sub_vars(sql, args, db)
      if not sql:
          raise ValueError(_('Report %(num)s has no SQL query.', num=id))
      self.log.debug('IG: Executing report with SQL "%s"' % sql)
      self.log.debug('IG: Request args: %r' % req.args)
      cursor = db.cursor()
      
      # The column name is obtained.
      get_col_name_sql = 'SELECT * FROM ( ' + sql + ' ) AS tab LIMIT 1'
      cursor.execute(get_col_name_sql, args)
      self.env.log.debug("IG: Query SQL(Get col names): " + get_col_name_sql)
      return cursor
    
    def _sql_columns(self, req, id, sql):
      r"""Retrieve the description of columns returned by a SQL 
      report.
      """
      repmdl = self.repmdl
      db = self.env.get_db_cnx()
      try:
        args = repmdl.get_var_args(req)
      except ValueError,e:
        raise ValueError(_('Report failed: %(error)s', error=e))
      try:
          cursor = self._sql_cursor(req, db, id, sql, args)
      except Exception, e:
          db.rollback()
          raise 
      else:
        self.log.debug('IG: Cursor desc %s', cursor.description)
        cols = list(get_column_desc(cursor, True))
        cursor.close()
        return cols
    
    def enum_columns(self, req, id):
        r"""Retrieve the columns present in a custom report.
        
        @param id     the report ID.
        @return       a list of tuples of the form 
                      (name, type, [label]).
        @throws       `NotImplementedError` if the report definition 
                      consists of saved custom query specified 
                      using a URL.
        @throws       `QuerySyntaxError` if the report definition 
                      consists of a `TracQuery` containing syntax errors.
        @throws       `Exception` in case of detecting any other error.
        """
        sql = self.get(req, id)['query']
        query = ''.join([line.strip() for line in sql.splitlines()])
        if query and (query[0] == '?' or query.startswith('query:?')):
          raise NotImplementedError('Saved custom queries specified ' \
                                  'using URLs are not supported.')
        elif query.startswith('query:'):
          query = Query.from_string(self.env, query[6:], report=id)
          fields = query.fields
          return [(f['name'], 'string', _(f['label'])) for f in fields] + \
                  [   ('changetime', 'number', _('Modified')), \
                      ('time', 'number', _('Created')), \
                      ('href', 'string', _('URL')), \
                      ('id', 'number', _('Ticket')), \
                  ]
        else:
          return self._sql_columns(req, id, sql)

#--------------------------------------------------
#   Version Control RPC
#--------------------------------------------------

def _normalize_timestamp(repos, req, timestamp, default=None):
  r"""Normalize datetime and revision numbers. Return only 
  datetime values.
  """
  if isinstance(timestamp, datetime):
    return timestamp
  elif isinstance(timestamp, xmlrpclib.DateTime):
    return rpc_to_datetime(timestamp, req)
  elif isinstance(default, (datetime, date, time)): # Return default
    return default
  else:
    return datetime.now(req.tz)

def _filter_revs(seq, repos, req, start, stop, full, \
                  accessor=None, log=None):
  r"""Filter values in seq so that only references to revisions 
  commited during a time interval be enumerated. 
  Items are binary tuples of the form `(revision id, changeset object)`.
  
  @param seq        original sequence to be filtered.
  @param repos      the repository managed by VCS.
  @param req        object containing information about the user 
                    requesting information in repository and more.
  @param start      boundary value. Revisions older than 
                    this value will not be retrieved. Dates 
                    and revision numbers are both accepted.
  @param stop       boundary value. Younger revisions 
                    will not be retrieved. Dates 
                    and revision numbers are both accepted.
  @param full       retrieve also the changeset object for revision.
  @param accesor    a function used to access the revision value 
                    stored in each item of the input sequence.
  @return           a sequence of tuples. The firts item is the 
                    element of the original sequence for which 
                    revision is in input time interval. The second is 
                    the changeset object or None (depending upon the 
                    value of `full` parameter)
  """
  if seq is not None:
    seq = iter(seq)
  if log is None:
    class VoidLogger:
      def __getattr__(self, attrnm):
        return lambda *args, **kwds: None
    log = VoidLogger()
  DO_NOTHING, DO_RETURN, DO_YIELD = xrange(3)
  load_chgset = True
  if isinstance(start, types.StringTypes) and \
      isinstance(stop, types.StringTypes):
    load_chgset = False
    if not repos.rev_older_than(stop, start):
      def cond(rev, chgset):
        if repos.rev_older_than(rev, start):
          return DO_RETURN
        elif repos.rev_older_than(stop, rev):
          return DO_NOTHING
        else:
          return DO_YIELD
    else:
      return                      # `start` committed after `stop`
  elif isinstance(start, types.StringTypes):
    if stop is None:
      load_chgset = False
      def cond(rev, chgset):
        # Process starts in youngest so there is no need to skip revisions
        return repos.rev_older_than(rev, start) and DO_RETURN or DO_YIELD
    else:
      ts = _normalize_timestamp(repos, req, stop)
      stop = repos.youngest_rev
      def cond(rev, chgset):
        if repos.rev_older_than(rev, start):
          return DO_RETURN
        else:
          if chgset.date < ts:
            return DO_YIELD
          else:
            return DO_NOTHING  
  elif isinstance(stop, types.StringTypes):
    if start is None:
      start = 0
      load_chgset = False
      def cond(rev, chgset):
        # Start in `stop` and stop in oldest so no need for DO_RETURN 
        # No need for DO_NOTHING since iterators will start from stop 
        # (assuming that revision numbers are valid and that's a 
        # precondition ;)
        return DO_YIELD
    else:
      ts = _normalize_timestamp(repos, req, start, _epoc)
      def cond(rev, chgset):
        # We start from `stop` so no need for DO_NOTHING ;)
        if chgset.date < ts:
            return DO_RETURN
        else:
          return DO_YIELD
  else:
    start_ts = _normalize_timestamp(repos, req, start, _epoc)
    stop_ts = _normalize_timestamp(repos, req, stop)
    start, stop = 0, repos.youngest_rev
    def cond(rev, chgset):
      ts = chgset.date
      if ts < start_ts:
          return DO_RETURN
      elif ts < stop_ts:
          return DO_YIELD
      else:
          return DO_NOTHING
  # Search backwards
  load_chgset = load_chgset or full
  try:
    while True:         # Stops when StopIteration is raised by seq
      item = seq.next()
      log.debug("IG: Processing item %s", item)
      if accessor:
        rev = accessor(item)
      else:
        rev = item
      log.debug("IG: Processing revision %s", rev)
      if repos.authz.has_permission_for_changeset(rev):
        try:
          chgset = load_chgset and repos.get_changeset(rev) or None
          action = cond(rev, chgset)
        except NoSuchChangeset:
          continue
        if action == DO_RETURN:
            return
        elif action == DO_YIELD:    # Implicit DO_NOTHING
          if full:
            yield item, chgset
          else:
            yield item, None
  except NoSuchChangeset:
    return

class VersionControlRPC(Component):
    r""" An interface to Trac's `Repository` and `RepositoryManager`.
    """ 
    implements(IXMLRPCHandler)

    # IXMLRPCHandler methods
    def xmlrpc_namespace(self):
        return 'source'
    
    def xmlrpc_methods(self):
        yield ('BROWSER_VIEW', 
                ((list, list, str, str, bool, int),
                 (list, list, str, str, bool), 
                 (list, list, str, str), 
                 (list, list, str), 
                 (list, list),), 
                 self.ls)
        yield ('BROWSER_VIEW', 
                ((list, list, str),
                 (list, list),), 
                 self.getFileAttributes)
        opt_types = [str, xmlrpclib.DateTime]
        yield ('CHANGESET_VIEW', 
                tuple(rpc_opt_sigs(list, None, opt_types, \
                                    opt_types, [bool])
                  ), 
                 self.getRevisions)
        yield ('CHANGESET_VIEW', 
                ((list, str, str, xmlrpclib.DateTime),
                 (list, str, str, str),
                 (list, str, str),
                 (list, str),), 
                 self.getFileHistory)
        yield ('CHANGESET_VIEW', 
                ((list, str),
                 (list, ),), 
                 self.enumChanges)
        yield ('CHANGESET_VIEW', 
                ((list, str),
                 (list, ),), 
                 self.normalize_rev) 
        yield ('CHANGESET_VIEW', 
                ((list, ),), 
                 self.quick_jump_entries) 
                 
    # Exported methods
    def ls(self, req, files, filter_by=None, rev=None, rec=False, depth=None):
        r"""List information about the FILEs. The root path makes 
        reference to the top-level folder configured for the 
        repository according to some options in `trac.ini`. File path 
        separator is always `/`.
        
        @param files      target files and|or folders
        @param filter_by  a UNIX filename pattern used to filter the 
                          results.
        @param rev        target revision number. If missing (or 
                          negative) it defaults to HEAD (youngest).
        @param rec        list files contained in this folder and 
                          its subfolders recursively.
        @param depth      if recursive mode is on, specify the maximum
                          recursion level. In this case use `0` for 
                          files in the same folder (i.e. no recursion), 
                          `1` for files in folder and its immediate 
                          child entries, `2`, `3` ... and so on. If 
                          this value is negative then full recursion 
                          is performed.
        @return           a list containing the full name (i.e. 
                          relative to the root folder in VCS) of all 
                          the files inside the input folders. File 
                          names are not generated in order so files 
                          in a folder and its subfolders may appear 
                          in different positions. If the revision ID 
                          supplied in is invalid then an empty list 
                          is returned.
        """
        repos = RepositoryManager(self.env).get_repository(req.authname)
        if rev is None: rev = repos.youngest_rev
        if depth < 0: depth = None
        no_depth = depth is None
        already = set()
        if filter_by is not None:
          filter_by = compile_pattern(filter_by)
        
        for item in files:
          if isinstance(item, types.StringTypes):
            try:
              d, node = 0, repos.get_node(item, rev)
              path = item
            except NoSuchNode:
              continue
            except NoSuchChangeset:
              return
          else:
            d, node = item
            path = node.path
          if path not in already: # Dont process filename twice
            if node.isfile:
              if (not filter_by) or filter_by.match(path) is not None:
                yield path
            elif node.isdir:
              for child in node.get_entries():
                if (not filter_by) or \
                              filter_by.match(child.path) is not None:
                  yield child.path
                if child.isdir and rec and (no_depth or d < depth):
                  files.append([d + 1, child])
                elif child.isfile:
                  already.add(child.path)     # Mark filename
                else:
                  self.log.error("Unknown node type %s at %s", \
                                                node.kind, node.path)
            else:
              self.log.error("Unknown node type %s at %s", \
                                                node.kind, node.path)
            already.add(path)                 # Mark filename
    
    def getFileAttributes(self, req, files, rev=None):
        r"""Retrieve the attributes of a group of files. The root 
        path makes reference to the top-level folder configured for 
        the repository according to some options in `trac.ini`. File 
        path separator is always `/`.
        
        @param files      target files and|or folders
        @param rev        target revision number. If missing (or 
                          negative) it defaults to HEAD (youngest).
        @return           a list of dictionaries. Each one of them 
                          contains the attributes of the input file 
                          at that position. All data is calculated 
                          with respect to the target revision (see 
                          `rev` parameter). If the file does not 
                          exist or was created after commiting the 
                          target revision then `None` is returned 
                          instead. The following attributes are 
                          supported:
                          
                          - path :    filename (full path from root 
                                      folder)
                          - kind :    the type of node (e.g. one of 
                                        'file' or 'dir') at `path`.
                          - size :    file size in bytes
                          - ext  :    file extension
                          - mime :    MIME type if known
                          - changed : Modification date (date of `lastrev`)
                          - lastrev : Revision (prior to the target 
                                      revision) when the latest 
                                      modifications to this file were 
                                      commited
                          - log :     Commit message for last revision 
                                      (`chgrev`)
        """
        repos = RepositoryManager(self.env).get_repository(req.authname)
        if rev is None:
          rev = repos.youngest_rev
        mimeview = Mimeview(self.env)
        changesets = {}
        for path in files:
          try:
            node = repos.get_node(path, rev)
          except NoSuchNode:
            yield {}
            continue
          _rev = node.rev
          attrs = dict(path=node.path, kind=node.kind, lastrev=_rev)
          try:
            chgset = changesets[_rev]
            attrs.update(dict(changed=chgset.date, log=chgset.message))
          except KeyError:
            try:
              changesets[_rev] = chgset = repos.get_changeset(_rev)
              attrs.update(dict(changed=chgset.date, log=chgset.message))
            except NoSuchChangeset:
              changesets[_rev] = attrs['changed'] = attrs['log'] = None
          if node.isdir:
            attrs.update(dict(size=0, ext='', mime=''))
          elif node.isfile:
            # Copycat ! from trac.versioncontrol.web_ui.browser
            # MIME type detection 
            content = node.get_content()
            _chunk = content.read(CHUNK_SIZE)
            mime_type = node.content_type
            if not mime_type or mime_type == 'application/octet-stream':
                mime_type = mimeview.get_mimetype(node.name, _chunk) or \
                            mime_type or 'text/plain'
            
            attrs.update(size=node.get_content_length(), \
                          ext=splitext(node.path)[-1][1:], \
                          mime=mime_type)
          else:
            self.log.error("Unknown node type %s at %s", \
                                              node.kind, node.path)
          yield attrs
    
    REV_ATTRS = ('rev', 'message', 'author', 'date')
    
    def getRevisions(self, req, since=None, until=None, full=False):
        r"""Retrieve information about all revisions in a time 
        interval. Younger revisions should appear first.
        
        @param since      boundary value. Revisions older than 
                          this value will not be retrieved. Dates 
                          and revision numbers are both accepted.
        @param until      boundary value. Younger revisions 
                          will not be retrieved. Dates 
                          and revision numbers are both accepted.
        @param full       level of detail used to generate the data. 
                          If false only revision numbers 
                          are returned, else tuples of the form 
                          (rev, message, author, date) are returned 
                          for each changeset, where :
                          
                          - rev     : is the revision number (ID)
                          - message : commit message provided by 
                                      Version Control System
                          - author  : the user that committed changes 
                                      in this revision.
                          - date    : the date when this revision was 
                                      commited (as determined by VCS).
        @return           a list of data representing each 
                          revision in interval. The kind of 
                          information returned is controlled by `full` 
                          parameter.
                          
                          Note: Some revisions may be skipped if 
                          permissions state that the user performing 
                          the request has no access to that 
                          particular changeset.
        """
        repos = RepositoryManager(self.env).get_repository(req.authname)
        last_cset = None
        try:
          if isinstance(until, types.StringTypes):
            last_cset = repos.get_changeset(until)
            until = last_cset.date
          elif isinstance(until, type(None)):
            until = datetime.now(req.tz)
          elif not isinstance(until, datetime):
            return []
          
          if isinstance(since, types.StringTypes):
            since = repos.get_changeset(since).date
          elif isinstance(since, type(None)):
            since = _epoc
          elif not isinstance(since, datetime):
            return []
        except NoSuchChangeset:
          return []
        else:
          def vcs_order(rev1, rev2):
            if rev1 == rev2:
              return 0
            rev1 = repos.normalize_rev(rev1)
            rev2 = repos.normalize_rev(rev2)
            if rev1 == rev2:
              return 0
            else:
              return repos.rev_older_than(rev1, rev2) and 1 or -1
          seq = list(repos.get_changesets(since, until))
          if last_cset:
            seq.append(last_cset)
          seq.sort(vcs_order, lambda cset: cset.rev)
          if full:
            return (tuple(getattr(chg, a) for a in self.REV_ATTRS) \
                      for chg in seq)
          else:
            return (chg.rev for chg in seq)
    
    def getFileHistory(self, req, path, rev=None, since=None):
        r"""Retrieve information about all the changes performed on a 
        file or directory in a time interval.
        
        @param path       file path in repository.
        @param since      boundary value. Revisions older than 
                          this value will not be retrieved. Dates 
                          and revision numbers are both accepted.
        @param rev        boundary value. Younger revisions 
                          will not be retrieved. Only revision 
                          numbers are allowed in this argument.
        @return           a list of `(path, rev, chg)` tuples, 
                          one for each revision in which the target 
                          was changed. This generator will follow 
                          copies and moves of a node (if the 
                          underlying version control system supports
                          that), which will be indicated by the 
                          first element of the tuple (i.e. the path) 
                          changing. Starts with an entry for the 
                          current revision.
                          
                          - path      : path the change was performed 
                                        upon
                          - rev       : revision number (ID)
                          - chg       : the kind of change being 
                                        performed. Supported values 
                                        are :
                            * add     : target was placed under 
                                        version control
                            * copy    : target was copied from 
                                        another location
                            * delete  : target was deleted
                            * edit    : target was modified
                            * move    : target was moved to 
                                        another location
                          
                          Note: Some revisions may be skipped if 
                          permissions state that the user performing 
                          the request has no access to that 
                          particular changeset.
        """
        if not isinstance(rev, (type(None), types.StringTypes)):
          raise ValueError("Revision ID must be a string")
        repos = RepositoryManager(self.env).get_repository(req.authname)
        try:
          node = repos.get_node(path, rev)
        except NoSuchNode:
          return []
        seq = node.get_history()
        seq = _filter_revs(seq, repos, req, since, rev, False, \
                            accessor=lambda x: x[1], log=self.log)
        return (x[0] for x in seq)
    
    def enumChanges(self, req, rev=None):
        r"""Enumerate all the changes performed at a given revision.
        
        @param rev        The ID of the target revision (changeset). 
                          Only numbers are allowed in this argument.
        @return           A list of tuples describing every change in 
                          the changeset. The tuple will contain 
                          `(path, kind, change, base_path, base_rev)`,
                          where :
                          
                          - change    : the kind of change being 
                                        performed (for further details 
                                        read the docstrings of 
                                        `getFileHistory` method).
                          - kind      : the type of node (e.g. one of 
                                        'file' or 'dir') at `path`.
                          - path      : is the targeted path for the 
                                        `change` (which is the 
                                        ''deleted'' path  for a 
                                        DELETE change).
                          - base_path : the source path for the
                                        action (empty string  in the 
                                        case of an ADD change).
                          - base_rev  : the source rev for the
                                        action (empty string in the 
                                        case of an ADD change).
                          
                          Note: Some revisions may be skipped if 
                          permissions state that the user performing 
                          the request has no access to that 
                          particular changeset.
        """
        repos = RepositoryManager(self.env).get_repository(req.authname)
        if rev is None:
          rev = repos.youngest_rev
        try:
          chgset = repos.get_changeset(rev)
        except NoSuchChangeset:
          return []
        return ((p, k, chg, bp or '', brev or '') \
                      for p, k, chg, bp, brev in chgset.get_changes())
    
    def normalize_rev(self, req, rev=None):
        r"""Return a canonical representation of a revision.

        It's up to the backend to decide which string values of `rev` 
        (usually provided by the user) should be accepted, and how they 
        should be normalized. Some backends may for instance want to 
        match against known tags or branch names.
        
        In addition, if `rev` is missing or '', the youngest revision 
        should be returned.
        
        @param rev        the ID of the target revision (changeset).
        @return           a revision ID as determined by the backend.
        """
        repos = RepositoryManager(self.env).get_repository(req.authname)
        return repos.normalize_rev(rev)
    
    def quick_jump_entries(self, req):
        r"""Retrieve a list of interesting places in the repository.
        
        @return           the generated results are of the form 
                          `(category, name, path, rev)` where:
                          
                          - category :  is one of `branches` or `tags`.
                          - name :      is the entry name.
                          - path :      the relevant path in the 
                                        repository.
                          - rev :       the target revision ID.
        """
        repos = RepositoryManager(self.env).get_repository(req.authname)
        return repos.get_quickjump_entries(None)
    



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Data sources used to publish the data about files managed by the 
Version Control System (e.g. `svn`) used by Trac. This includes 
revisions, change log, and (source) files.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from api import gviz_col, gviz_param, GVizBadRequestError
from util import GVizXMLRPCAdapter, convert_req_date, \
                  REQFIELDS_DESC, REQFIELDS_DEFAULTS

from itertools import chain
import types

#--------------------------------------------------
# Source files
#--------------------------------------------------

class GVizSourceBase(GVizXMLRPCAdapter):
  abstract = True
  
  FA_CODE = compile('(path, kind, size, ext, mime, changed, lastrev)', \
                            '<string>', 'eval')
  
  def _normalize_rev(self, rev, req, fmt=REQFIELDS_DEFAULTS['datefmt']):
    r"""Converts a value in a client request making reference to a 
    revision committed to the source repository.
    
    @param rev          the value received in the HTTP request.
    @param fmt          datetime format used to convert string (not 
                        needed if allow_date is false).
    @param req          request object (not needed if allow_date is 
                        false, mandatory otherwise).
    """
    try:
      return convert_req_date(rev, fmt, req, False)
    except GVizBadRequestError:
      return rev
  
  def _requested_files(self, req, files, filter=None, rev=None, \
                        rec=False, depth=None):
    if isinstance(files, types.StringTypes):
      files = [files]
    if depth is not None:
      try:
        depth = int(depth)
      except ValueError:
        raise GVizBadRequestError("Recursion depth must be an integer value")
    return self._rpc_obj.ls(req, files, filter, rev, rec, depth)
    
  def xmlrpc_namespace(self):
      r"""Use Version Control XML-RPC.
      """
      return ('source',)

class GVizSourceFiles(GVizSourceBase):
    r"""Return information about the files managed by the Version 
    Control System (e.g. `svn`) connected to the Trac environment.
    
    This component depends on tracgviz.rpc.VersionControlRPC. The 
    later must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    
    # IGVizDataProvider methods
    def get_data_schema(self, req):
        r"""Return file attributes.
        """
        return [('path', 'string'), ('kind', 'string'), \
                ('size', 'number'), \
                ('ext', 'string'), ('mime', 'string'), \
                ('changed', 'datetime'), ('lastrev', 'string'),
                ('log', 'string'), ]
    
    @gviz_param('file',     "target files and|or folders. Can be "
                              "specified more than once in order to "
                              "process multiple files.")
    @gviz_param('filter', "a UNIX filename pattern used to filter "
                      "the results.")
    @gviz_param('rev',       "target revision number. If missing (or "
                      "negative) it defaults to HEAD (youngest).")
    @gviz_param('rec',       "list files contained in this folder and "
                      "its subfolders recursively.")
    @gviz_param('depth',     "if recursive mode is on, specify the "
                      "maximum recursion level. In this case use `0` "
                      "for files in the same folder (i.e. no recursion), "
                      "`1` for files in folder and its immediate "
                      "child entries, `2`, `3` ... and so on. If "
                      "this value is negative then full recursion "
                      "is performed.")
    @gviz_col('path',    "filename (full path from root "
                          "folder)")
    @gviz_col('kind',     "the type of node (e.g. one of "
                          "'file' or 'dir') at `path`.")
    @gviz_col('size',    "file size in bytes")
    @gviz_col('ext', "    file extension")
    @gviz_col('mime',    "MIME type if known")
    @gviz_col('changed', "Modification date (date of `lastrev`)")
    @gviz_col('lastrev', "Revision (prior to the target "
                          "revision) when the latest "
                          "modifications to this file were "
                          "commited")
    @gviz_col('log',     "Commit message for last revision "
                          "(`lastrev`)")
    def get_data(self, req, tq, file='/', filter=None, rev=None, \
                        rec=False, depth=None, **tqx):
        r"""Retrieve files attributes.
        """
        files = self._requested_files(req, file, filter, rev, rec, \
                                        depth)
        if rec is not False:
          rec = True
        revs = dict()
        
        fa_meth = self._rpc_obj.getFileAttributes
        ra_meth = self._rpc_obj.getRevisions
        return (eval(self.FA_CODE, attrs) + \
                    (revs.setdefault(attrs['lastrev'], \
                                      iter(ra_meth(req, \
                                              str(attrs['lastrev']), \
                                              str(attrs['lastrev']), \
                                              True)).next()[1]),) \
                    for attrs in fa_meth(req, files, rev) if attrs)
    
    def gviz_namespace(self):
        return ('source', 'files')

class GVizSourceFileHistory(GVizSourceBase):
    r"""Return information about the historical events recorded by 
    Version Control System (e.g. `svn`) leading to the presence 
    of a group of files in a given revision.
    
    This component depends on tracgviz.rpc.VersionControlRPC. The 
    later must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    
    # IGVizDataProvider methods
    def get_data_schema(self, req):
        r"""Return event fields.
        """
        return [('path', 'string'), ('target', 'string'), \
                ('rev', 'string'), ('chg', 'string'),
                ('kind', 'string'), ('size', 'number'), \
                ('ext', 'string'), ('mime', 'string'), \
                ('changed', 'datetime'), ('log', 'string'), 
                ]
    
    @gviz_param('file',     "target files and|or folders. Can be "
                              "specified more than once in order to "
                              "process multiple files.")
    @gviz_param('filter', "a UNIX filename pattern used to filter "
                      "the results.")
    @gviz_param('rev',       "target revision number. If missing (or "
                      "negative) it defaults to HEAD (youngest).")
    @gviz_param('rec',       "list files contained in this folder and "
                      "its subfolders recursively.")
    @gviz_param('depth',     "if recursive mode is on, specify the "
                      "maximum recursion level. In this case use `0` "
                      "for files in the same folder (i.e. no recursion), "
                      "`1` for files in folder and its immediate "
                      "child entries, `2`, `3` ... and so on. If "
                      "this value is negative then full recursion "
                      "is performed.")
    @gviz_param('since',     "boundary value. Revisions older than "
                          "this value will not be retrieved. Dates "
                          "and revision numbers are both accepted.")
    @gviz_param('fmt', REQFIELDS_DESC['datefmt'] % \
                            dict(args='`since`', plural=''))
    @gviz_col('path',     "this field makes reference to the original "
                          "path of the file whose changes have "
                          "been requested.")
    @gviz_col('target',   "path to the file upon which the change "
                          "was performed.")
    @gviz_col('rev',      "revision number (ID).")
    @gviz_col('chg',      "the kind of change being "
                          "performed. Supported values "
                          "are :"
                          "  * add     : target was placed under "
                                      "version control"
                          "  * copy    : target was copied from "
                                      "another location"
                          "  * delete  : target was deleted"
                          "  * edit    : target was modified"
                          "  * move    : target was moved to "
                                      "another location")
    @gviz_col('kind',     "the type of node (e.g. one of "
                          "'file' or 'dir') at `target`.")
    @gviz_col('size',    "size of `target` in bytes")
    @gviz_col('ext',     "extension of `target`")
    @gviz_col('mime',    "MIME type of `target` if known")
    @gviz_col('changed', "Modification date of `target` (date of `rev`)")
    @gviz_col('log',     "Commit message for last revision "
                          "(`lastrev`)")
    def get_data(self, req, tq, file='/', filter=None, rev=None, \
                        rec=False, depth=None, since=None, \
                        fmt=REQFIELDS_DEFAULTS['datefmt'], **tqx):
        r"""Retrieve changes in file history.
        """
        files = self._requested_files(req, file, filter, rev, rec, \
                                        depth)
        if since is not None:
          since = self._normalize_rev(since, req, fmt)
        fh_meth = self._rpc_obj.getFileHistory
        fa_meth = self._rpc_obj.getFileAttributes
        ra_meth = self._rpc_obj.getRevisions
        revs = dict()
        def_attrs = {'size' : None, 'mime' : None, 'changed' : None,
                      'kind' : None, 'last_rev': None}
        for path in files:
          for target, _rev, chg in fh_meth(req, path, rev, since):
            attrs = fa_meth(req, [target], _rev)[0] or \
                    def_attrs.copy().update(dict(path=target, \
                                          ext=splitext(target)[-1][1:]))
            yield (path, target, _rev, chg) + \
                    eval(self.FA_CODE, attrs)[1:-1] + \
                    (revs.setdefault(_rev, \
                                      iter(ra_meth(req, str(_rev), 
                                            str(_rev), True)).next()[1]),)
    
    def gviz_namespace(self):
        return ('source', 'file', 'history')

#--------------------------------------------------
# Chagesets
#--------------------------------------------------

class GVizChangesetDetails(GVizSourceBase):
    r"""Return a set of changes committed at once in a repository.
    
    This component depends on tracgviz.rpc.VersionControlRPC. The 
    later must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    
    # IGVizDataProvider methods
    def get_data_schema(self, req):
        r"""Return event fields.
        """
        return [('rev', 'string'), ('chg', 'string'), \
                ('kind', 'string'), ('target', 'string'), \
                ('base_path', 'string'), ('base_rev', 'string') \
                ]
    
    @gviz_param('rev',       "target revision number. If missing (or "
                      "negative) it defaults to HEAD (youngest). "
                      "This parameter can be specified more than once "
                      "in order to retrieve changes committed in "
                      "multiple revisions.")
    @gviz_param('since',    "used to select revisions in an interval. "
                            "Works together with `until` parameter, "
                            "but nothing happens if both are missing. "
                            "Revisions younger than "
                            "this value and older than `until` will "
                            "be processed too. Dates "
                            "and revision numbers are both accepted.")
    @gviz_param('until',    "used to select revisions in an interval. "
                            "Works together with `since` parameter, "
                            "but nothing happens if both are missing. "
                            "Revisions older than "
                            "this value and younger than `since` will "
                            "be processed too. Dates "
                            "and revision numbers are both accepted.")
    @gviz_param('fmt', REQFIELDS_DESC['datefmt'] % \
                            dict(args='both `since` and `until`', \
                                  plural='s'))
    @gviz_col('rev',      "revision number (ID).")
    @gviz_col('chg',      "the kind of change being "
                          "performed. Supported values "
                          "are :"
                          "  * add     : target was placed under "
                                      "version control\n"
                          "  * copy    : target was copied from "
                                      "another location\n"
                          "  * delete  : target was deleted\n"
                          "  * edit    : target was modified\n"
                          "  * move    : target was moved to "
                                      "another location")
    @gviz_col('kind',     "the type of node (e.g. one of "
                          "'file' or 'dir') at `path`.")
    @gviz_col('target',   "path to the file involved in the change.")
    @gviz_col('base_path',"the source path for the"
                          "action (`None` in the case of "
                          "an ADD change).")
    @gviz_col('base_rev', "the source rev for the"
                          "action (`-1` in the case of "
                          "an ADD change).")
    def get_data(self, req, tq, rev=None, since=None, until=None, 
                                fmt=REQFIELDS_DEFAULTS['datefmt'], **tqx):
        r"""Retrieve changes in revision.
        """
        rpc_obj = self._rpc_obj
        already = set()
        if not isinstance(rev, list):
          rev = (rev,)
        for r in rev:
          try:
            r = rpc_obj.normalize_rev(req, r)
          except:
            raise
          if r not in already:
            for chg in rpc_obj.enumChanges(req, r):
              yield (r,) + chg
            already.add(r)       # Mark revision
        if (since, until) != (None, None):
          if since is not None:
            since = self._normalize_rev(since, req, fmt)
          if until is not None:
            until = self._normalize_rev(until, req, fmt)
          for r in rpc_obj.getRevisions(req, since, until):
            if r not in already:
              for chg in rpc_obj.enumChanges(req, r):
                yield (r,) + chg
    
    def gviz_namespace(self):
        return ('source', 'changes')

class GVizChangesets(GVizSourceBase):
    r"""Return meta-data about all the revisions commited to the 
    repository.
    
    This component depends on tracgviz.rpc.VersionControlRPC. The 
    later must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    
    # IGVizDataProvider methods
    def get_data_schema(self, req):
        r"""Return fields in revision log.
        """
        return [('rev', 'string'), ('message', 'string'), \
                 ('author', 'string'), ('date', 'datetime')]
    
    @gviz_param('rev',       "target revision number. If missing (or "
                      "negative) it defaults to HEAD (youngest). "
                      "This parameter can be specified more than once "
                      "in order to retrieve data about multiple "
                      "revisions.")
    @gviz_param('since',    "boundary value. Revisions older than "
                            "this value will not be retrieved. Dates "
                            "and revision numbers are both accepted.")
    @gviz_param('until',    "boundary value. Younger revisions "
                            "will not be retrieved. Dates "
                            "and revision numbers are both accepted.")
    @gviz_param('fmt', REQFIELDS_DESC['datefmt'] % \
                            dict(args='both `since` and `until`', \
                                  plural='s'))
    @gviz_col('rev',        "is the revision number (ID)")
    @gviz_col('message',    "commit message provided by "
                            "Version Control System")
    @gviz_col('author',     "the user that committed changes "
                            "in this revision.")
    @gviz_col('date',       "the moment when this revision was "
                            "commited (as determined by VCS).")
    def get_data(self, req, tq, rev=None, since=None, until=None, 
                        fmt=REQFIELDS_DEFAULTS['datefmt'], **tqx):
        r"""Retrieve revision log.
        """
        gr_meth = self._rpc_obj.getRevisions
        if rev is not None:
          if isinstance(rev, types.StringTypes):
            rev = (rev,)
          already = set(self._rpc_obj.normalize_rev(req, r) for r in rev)
          revlog = (gr_meth(req, str(r), str(r), True)[0] for r in already)
        else:
          revlog = already = []
	if rev is None or since is not None or until is not None:
          if since is not None:
            since = self._normalize_rev(since, req, fmt)
          if until is not None:
            until = self._normalize_rev(until, req, fmt)
          return chain(revlog, 
                        (ra for ra in gr_meth(req, since, until, True) \
                                    if ra[0] not in already))
	else:
	  return revlog
    
    def gviz_namespace(self):
        return ('source', 'revlog')

class GVizQuickJumpEntries(GVizSourceBase):
    r"""Retrieve interesting places in the repository (a.k.a. 
    `quick jump entries`).
    
    This component depends on tracgviz.rpc.VersionControlRPC. The 
    later must be enabled. Please read 
    https://opensvn.csie.org/traccgi/swlcu/wiki/En/Devel/TracGViz/DataSources#Preconditions
    for further details.
    """
    
    # IGVizDataProvider methods
    def get_data_schema(self, req):
        r"""Return details about quick jump entries.
        """
        return [('category', 'string'), ('name', 'string'), \
                 ('path', 'string'), ('rev', 'string')]
    
    @gviz_col('category',   "is one of `branches` or `tags`.")
    @gviz_col('name',       "is the entry name.")
    @gviz_col('path',       "the relevant path in the repository.")
    @gviz_col('rev',        "the target revision ID.")
    def get_data(self, req, tq, **tqx):
        r"""Retrieve quick jump entries.
        """
        return self._rpc_obj.quick_jump_entries(req)
    
    def gviz_namespace(self):
        return ('source', 'labels')




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Core components to integrate Trac with Google Gadgets technology
(i.e. iGoogle container). 

This package states how to enhance and extend the different features
available in Trac Gadgets subsystem. The goal for all this 
extensibility is to ease the process of embedding gadgets in Trac 
site by providing easy-to-use wiki formatting constructs.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from trac.core import Interface, Component, ExtensionPoint, implements

class IGadgetBorderImgStore(Interface):
    r"""The interface implemented by the components responsible for
    storing the pictures to use in gadgets borders, and keep track of
    the border sets.
    
    Images are stored in a directory inside the environment's 
    `htdocs/gadget/border` folder. This dir name matchs the name 
    chosen for the border id. Therefore the URL specified in the 
    gadget are available at URLs like :
    
    - http://<server>/<env>/chrome/site/gadget/border/<border_id>/
    
    ... or using any of the following TracLinks syntax:
    
    - htdocs:gadget/border/<border_id>/
    
    - gadget:border:<border_id>
    
    The latest is the most portable way (and therefore is 
    recommended) since images might be stored in a different
    repository.
    
    Note: The final slash MUST always be there, except in 
            TracLinks starting with `gadget:border` prefix.
    """
    def store_images(id, tl, tr, bl, br, l, r, b, tt, tn):
        """Store the images in an image-based gadget border.
        
        @param id the border identifier
        @param tl the contents of the top-left corner image
        @param tr the contents of the top-right corner image
        @param bl the contents of the bottom-left corner image
        @param br the contents of the bottom-right corner image
        @param l the contents of the left side image
        @param r the contents of the right side image
        @param b  the contents of the bottom image
        @param tt the contents of the top level image (with title)
        @param tn  the contents of the top level image (no title)
        
        Note: All the previous values (except `id`) may also be file 
                objects from where contents will be retrieved.
        """
    def load_image(id, img_id):
        r"""Retrieve the contents of a picture in an image-based 
        gadget border.
        
        @param id the border identifier
        @param img_id one of `tl`, `tr`, `bl`, `br`, `l`, `r`, `b`, 
                    `tt`, `tn` so as to identify the requested image
        @return a file-like object used to retrieve the image contents
                    or `None` if such file does not exists
        """

BORDER_IMAGES = ['tl', 'tr', 'bl', 'br', 'l', 'r', 'b', 'tt', 'tn']

class IGadgetBorderProvider(Interface):
    r"""Interface implemented by all those components offering borders
    gadgets will be surronded with. This includes both image-based 
    and CSS-based borders.
    """
    def get_namespace():
        r"""Specify the namespace identifying this borders provider.
        
        @return a string or sequence of strings
        """
    def get_border_spec(border_id):
        r"""Retrieve the string to set the `border` parameter in the 
        URL needed to paste the gadget into the HTML of the target 
        web page.
        
        @param border_id a string identifying the target border
        @return a string value formatted as explained for gatget 
                borders in iGoogle Gadgets specification.
                See code.google.com/apis/gadgets/docs/legacy/publish.html#Borders
                for more details.
        """
    def get_description():
        r"""Describe this border provider.
        
        @return an informative string describing this border provider.
        """
    def enum_borders():
        r"""List all the borders managed by this particular 
        border provider.
        
        @return a sequence containing all the available border ids
        """

class GadgetBordersManager(Component):
    r"""Component responsible for managing gadget borders.
    """
    bprov = ExtensionPoint(IGadgetBorderProvider)
    
    # TODO : Implement
    def __init__(self):
        self.providers = dict([p.get_namespace(), p] for p in self.bprov)
    




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

r"""Trac Data Source able to feed widgets implemented with Google Visualization API.

Components allowing Trac to export a project (environment)'s data so
that different widgets based on Google Visualization (GViz) API
be able to use that information in order to render a web GUI item.

Copyright 2009-2011 Olemis Lang <olemis at gmail.com>
Licensed under the Apache License, Version 2.0 
"""

from trac.core import Interface, Component, ExtensionPoint, \
                      implements
from trac.perm import IPermissionRequestor, PermissionError
from trac.util.datefmt import utc
from trac.web.api import IRequestHandler, RequestDone
from trac.web.chrome import ITemplateProvider

from pkg_resources import resource_filename

try:
    import gviz_api
except ImportError:
    import _gviz_api as gviz_api
import sys
from xmlrpclib import DateTime
from datetime import datetime

class GVizException(Exception):
  r"""Base class for all exception types defined in this package"""

class GVizProviderNotFoundError(GVizException):
    r"""Exception raised to denote that there is no GViz provider able
    to handle the request.
    """

class GVizNotSupportedError(GVizException):
    r"""Exception raised to denote that an unsopported feature has been
    requested by the client.
    """

class GVizBadRequestError(GVizException):
    r"""Exception raised to denote that the client request contains 
    wrong fields and/or values, or that required data is missing.
    """

class GVizUnknownProviderError(GVizException):
    r"""Exception raised to denote that there is no provider able to
    to handle the client request.
    """

class GVizNotAuthenticatedError(GVizException):
    r"""Exception raised when a permission error is detected while 
    processing an anonymous request.
    """

class GVizDataNotModifiedError(GVizException):
    r"""Exception raised when the hash generated from the data
    returned by the GViz provider matches the hash value sent by the 
    client.
    """
    message = ''

class GVizInvalidConfigError(GVizException):
    r"""Exception raised when an invalid, incorrect or unsupported 
    value has been assigned to a configuration option.
    """

def gviz_col(col_id, col_doc):
    r"""This function can be used to document the meaning of the 
    different columns in the table returned by GViz data sources as 
    well as the default label and column order. This function is 
    compliant with annotations as defined by PEP 3107.
    """
    def decorator(func):
        try:
            anns = func.func_annotations
        except AttributeError:
            func.func_annotations = {'return' : \
                    {col_id: col_doc}}
        else:
            anns.setdefault('return', {})[col_id] = col_doc
        return func
    return decorator

def gviz_param(name, doc):
    r"""This function can be used to document the meaning of the 
    different parameters accepted by GViz data sources as 
    well as their default values. 
    This function is compliant with annotations as defined by PEP 3107.
    """
    if name == 'return':
        return lambda f: f
    def decorator(func):
        try:
            anns = func.func_annotations
        except AttributeError:
            func.func_annotations = {name : doc}
        else:
            anns[name] = doc
        return func
    return decorator

class IGVizDataProvider(Interface):
    r"""All the components providing GViz data tables to clients
    ought to implement this interface.
    """
    def gviz_namespace():
        r""" Provide the path to access this data source provider.
        This is used to determine the URL clients will retrieve
        the data from.
        
        Return a list of items, each one being a string denoting
        either a path handle or a regular expression used to 
        match the requested path. Each such item *must not* match a
        string containing the slash (i.e. `/`) character.
        
        Note : regex in path handles are not supported yet.
        """
    
    def get_data_schema(req=None):
        r"""Provide the schema used to populate GViz data tables out 
        of the Python object containing the table contents.
        
        @req    (since version 1.3.3) an optional argument that can 
                be used by data providers offering data in different 
                ways depending upon the values provided at run-time 
                for some parameters. If this function has more than 
                a single argument then it *MUST* be the request 
                being processed.
        @return schema definition used to prepare the data table. See 
                the documentation for `gviz_api.DataTable` class for 
                more details about schemas.
        """
    
    def get_data(req, tq, **tqx):
        r""" Return a Python object containing all the data to be
        returned to the client so as to fill a data table object.
        
        @param tq the query requested by the client.
        @param req the query requested by the client.
        @param tqx custom arguments specified by the client. Every 
                    standard parameter defined by the Visualization
                    protocol (see the values in
                    `TracGVizSystem.reserved_args`) is removed since
                    they have special meanings no matter what data
                    is provided by a specific GViz data source.
        """

class IGVizProtocolHandler(Interface):
    r"""All components implementing a specific version of the Google
    Visualization Protocol must implement this interfaces.
    """
    def get_version():
        r""" Return a tuple like (epoch, major[, minor]) representing
        the protocol version implemented by the component.
        """
    
    def get_std_params():
        r"""Return a mapping object representing the parameters having
        special meanings according to the protocol specification.
        The mapping keys should be the name of the parameter, whereas
        the mapping values should be the parameter's default value.
        """
    
    def output_response(_req, _table, _error=None,
                        _warnings=None, **std_params):
        r"""Send the response back to the client.
        
        @param req an object encapsulating the data submitted by
                    the client in the HTTP request.
        @param table the instance of `gviz_api.DataTable` containing
                    the data requested by the client.
        @param error an exception object raised while processing the
                    request or `None` otherwise.
        @param warnings a sequence of warning messages returned to 
                    the client.
        @param std_params the values bound to the standard parameters
                    in the request string.
        """

class DataTable(gviz_api.DataTable):
    r"""Tries to fix the incompatibilities between xmlrpclib and 
    gviz_api modules.
    """
    @staticmethod
    def SingleValueToJS(value, value_type):
        r"""Translates a single value and type into a JS value, but 
        trying to fix the incompatibilities between xmlrpclib and 
        gviz_api modules.
        """
        if value_type in ["date", "datetime", "timeofday"] and \
                isinstance(value, DateTime):
            value = datetime.strptime(value.value, '%Y%m%dT%H:%M:%S')
        if value_type in ["date", "datetime", "timeofday"] and \
                isinstance(value, int):
            value = datetime.fromtimestamp(int(value or 0), utc)
        elif value_type == "string" and isinstance(value, unicode):
            value = value.encode('utf-8', 'replace')
        return gviz_api.DataTable.SingleValueToJS(value, value_type)

class TracGVizSystem(Component):
    r"""A component responsible for dispatching requests addressed to 
    the different GViz data sources available in the system. The 
    requests processed by the appropriate provider return a response 
    back to the client conforming to the formats supported by Google 
    Visualization API. By default it should return data in 
    `JSON Response Format`, but this decision is delegated to the 
    underlying protocol implementation.
    """
    implements(IRequestHandler, ITemplateProvider, \
                IPermissionRequestor)
    providers = ExtensionPoint(IGVizDataProvider)
    handlers = ExtensionPoint(IGVizProtocolHandler)
    
    def __init__(self):
        self._cache = dict(['/gviz/' + '/'.join(p.gviz_namespace()), p] \
                        for p in self.providers)
        self.log.debug('IG: Providers cache %s', (self._cache,))
        self._handlers = dict( \
                ['.'.join(str(x) for x in h.get_version()), h] \
                for h in self.handlers)
        try:
            self._latest = max(h.get_version() for h in self.handlers)
        except ValueError:
            self._latest = None
        else:
            self._latest = '.'.join(str(x) for x in self._latest)
    
    # IPermissionRequestor methods
    def get_permission_actions(self):
      """Return a list of actions defined by this component."""
      yield 'GVIZ_VIEW'
    
    # IRequestHandler methods
    def match_request(self, req):
        r"""Return whether the requested path starts with `/gviz`
        prefix or not.
        """
        url_path = req.path_info
        if url_path.startswith('/gviz'):
            try:
                return url_path[5] == '/'
            except IndexError:
                return True
    
    def process_request(self, req):
        r"""Perform one of the following actions :
        
        - If the request is addressed to a specific data source then
          send back the corresponding data table's contents in one of
          the formats supported by Google Visualization API (defaults
          to JSON Response Format).
        
        - If the request is addressed directly to the `root path` (i.e. 
          req.path_info == '/gviz') then render an informative page
          displaying all the registered data sources' specification.
        
        - Else send back a `Not Found` error message.
        """
        url_path = req.path_info
        if url_path in ('/gviz', '/gviz/'):
            req.perm.require('GVIZ_VIEW')
            return 'gviz_index.html', dict(), None
        else:
            if self._latest is None:
                req.send_response(404)
                req.end_headers()
                req.write('Unable to find any protocol handler')
                raise RequestDone()
            try:
                req.perm.require('GVIZ_VIEW')
                provider = self._cache[url_path]
            except (KeyError, PermissionError), exc:
                if isinstance(exc, KeyError):
                  exc = GVizUnknownProviderError('Unknown provider "%s"' %
                                              (url_path,))
                handler = self._handlers[self._latest]
                handler.output_response(req, None, exc, None)
                raise RequestDone()
            else:
                tq = req.args.pop('tq', '')
                try:
                    params = req.args['tqx']
                    self.log.debug("IG: Plain params : %s" % (params,))
                except KeyError:
                    std_params = dict()
                    handler = self._handlers[self._latest]
                else:
                    try:
                        params = dict(i.split(':', 1) \
                                for i in str(params).split(';'))
                    except:
                        exc = GVizBadRequestError('Syntax error in "%s"' %
                                            (params,))
                        handler = self._handlers[self._latest]
                        handler.output_response(req, None, exc, \
                                                None, **std_params)
                        raise RequestDone()
                    else:
                        version = params.get('version', self._latest)
                        try:
                            handler = self._handlers[version]
                            # TODO: Provide a better match for protocol handlers
                        except KeyError, exc:
                            handler = self._handlers[self._latest]
                            exc = GVizNotSupportedError('Unsupported protocol version '
                                    '"%s"' % (exc.message,))
                            handler.output_response(req, None, exc, \
                                                    None, **std_params)
                            raise RequestDone()
                        defaults = handler.get_std_params()
                        std_params = dict([k, params.pop(k, v)] \
                                for k,v in defaults.iteritems())
                # Custom options have to be specified outside
                # tqx field (i.e. directly inside the URL), 
                # at least until further notice
                self.log.debug("IG: Request parameters : %s", req.args)
                params = dict(req.args)
                self.log.debug("IG: Custom parameters : %s", params)
                table = None
                try:
                    data = provider.get_data(req, tq, **params)
                    sch = provider.get_data_schema.im_func.func_code
                    sch_args = (sch.co_argcount > 1) and (req,) or ()
                    table = DataTable(provider.get_data_schema(*sch_args), \
                                        data)
                    handler.output_response(req, table, None, \
                                            None, **std_params)
                except RequestDone:
                    raise
                except Exception, exc:
                    self.log.exception("IG: Handling exception")
                    handler.output_response(req, table, 
                            sys.exc_info()[1:], exc, **std_params)
                    raise RequestDone()
                
                raise RequestDone()
    
    # ITemplateProvider methods
    def get_htdocs_dirs(self):
        yield ('gviz', resource_filename('tracgviz', 'htdocs'))

    def get_templates_dirs(self):
        return [resource_filename('tracgviz', 'templates')]

class IGVizTableEncoder(Interface):
    r"""Implementing this interface is mandatory for every component
    in charge of converting a table contents into a character
    stream suitable for trasmitting it over an HTTP connection.
    """
    def supported_versions():
        r"""Return a sequence containing tuples, each one specifying
        the protocol versions supported by this format encoder. The
        tuples should be as follows :
        
        (Relationship, VersionNumber)
        
        where :
        
        Relationship := '<' | '>' | '<=' | '>=' | '==' | '!='
        VersionNumber := (Number ',' Number [',' Number])
        
        e.g. [('==', (0, 1, 2)),('==', (0, 1, 5)), 
              ('>=', (0, 2)), ('!=', (0, 4)), ('<', (1, 0))]
        """
    
    def stream_contents(table, params={}):
        r"""Convert the table contents to a format suitable for
        trasmitting them over an HTTP connection.
        
        @param table the target instance of `gviz_api.DataTable`.
        @std_params the values of the standard parameters.
        """
    
    def get_format_id():
        r"""Return the token identifying the format outputted by this 
        formatter object.
        """
    
    def get_content_type():
        r"""Return the content-type associated with this format 
        encoder.
        """

class ITracLinksHandler(Interface):
    r"""Interface implemented by those classes being
    TracLinks providers registered under the namespaces managed by
    this plugin.
    """
    def get_link_resolvers(self):
        r"""Return an iterable over (namespace, formatter) tuples.
        Namespaces value must be tuples.
        
        Each formatter should be a function of the form
        fmt(formatter, ns, target, label), and should
        return some HTML fragment.
        The `label` is already HTML escaped, whereas the `target` is not.
        """

class IHashLibrary(Interface):
  r"""Interface implemented by all those components defining one or 
  more secure hash, message digest, or checksum algorithms to be used 
  by other components.
  """
  def get_hash_properties(method_name):
    r"""Determine whether this component supports a secure hash 
    algorithm or not.
    
    @param method_name  is the name identifying the hash method
                        (e.g. sha1, md5, ...) in lowercase letters
    @return             `None` if the component does not support the 
                        requested algorithm. Otherwise, a tuple of 
                        the form (priority, source) where
                        
                        priority: is a number indicating how much 
                                  relevant is this implementation. 
                                  If multiple components support the 
                                  same method then the one specifying 
                                  the higher `priority` value will 
                                  be chosen.
                        source:   is a numeric constant identifying 
                                  the library providing the 
                                  implementation for this method. The 
                                  following values have been defined 
                                  so far:
                                  
                                  0   - implemented by the OpenSSL 
                                        library that Python uses on 
                                        your platform. 
                                  100 - implemented in Python stdlib
                                  200 - implemented in a Python module 
                                        (or C extension, ...) but not 
                                        in stdlib
    """
  def new_hash_obj(method_name, init_data):
    r"""Create a hash object.
    
    @param method_name  is the name identifying the hash method
                        (e.g. SHA1, MD5, ...)
    @param init_data    data processed since the beginning
    @return             a hash object implementing the aforementioned 
                        algorithm by means of the same 
                        simple interface defined in `hashlib` module.
    """

# TODO : Cache the requests and optimize the call/response life cycle
#            by using the `reqId` parameter.

